import importlib
from abc import ABC
from typing import Union

from how_about_no.core.logger import Logger
from how_about_no.core.vcs import BaseVCS


class BasePlugin(ABC):
    """A base class for the all plugins.

    A plugin can be used in two different ways: checking whether or not the builds
    should be skipped and populating that information further so it can be used in
    external system, eg. CI/CD.

    A basic implementation of a sample plugin can look like this:

        class SamplePlugin(BasePlugin):

            def check(self) -> bool:
                return False

            def populate(self, value: bool) -> None:
                print(value)

    Please note that the engine will gather the results of `check` methods of all
    plugins before deciding whether or not the builds can be skipped, so the value
    passed to method `populate` does not have to be the same as the value returned from
    method `check`.
    """

    def __init__(self, main_branch_name: str, vcs: BaseVCS, **kwargs) -> None:
        """Initialize the plugin.

        Child classes should override this method to validate the configuration or set
        default values for selected configuration options.

        Arguments:
            main_branch_name (str): The name of the main branch.
            vcs (BaseVCS): The Version Control System that should be used in the plugin.
            kwargs: Holds the plugin configuration loaded from the configuration file.
        """
        for key, value in kwargs.items():
            setattr(self, key, value)

        self.main_branch_name = main_branch_name
        self.vcs = vcs

        self._logger = Logger.get_logger()

    def check(self) -> bool:
        """Indicate whether or not the builds can be skipped.

        It returns `False` by default as it may not be implemented in a child class, and
        it does not make sense to cancel the builds because of that.

        Returns:
            (bool)
        """
        raise NotImplementedError()

    def populate(self, value: bool) -> None:  # pylint: disable=unused-argument
        """Populate the result of all checks further."""
        raise NotImplementedError()


def load_plugin(
    plugin_name: str, main_branch_name: str, vcs: BaseVCS, **kwargs
) -> Union[BasePlugin, None]:
    """Load the plugin with given name.

    Arguments:
        plugin_name (str): The name of the plugin to import.
        main_branch_name (str): The name of the main branch.
        vcs (BaseVCS): The Version Control System that should be used in the plugin.
        kwargs: Holds the plugin configuration loaded from the configuration file.

    Returns:
        (Union[BasePlugin, None]): The plugin class or None if not loaded for any
            reason.
    """
    try:
        plugin_class = importlib.import_module(plugin_name).PLUGIN
        return plugin_class(main_branch_name=main_branch_name, vcs=vcs, **kwargs)
    except (AttributeError, ImportError, TypeError):
        return None



from how_about_no.tests.fakes import FakePlugin, FakeVCS


def test_plugin_initialization():
    """
    Given a VCS, branch name and one more custom field
    When we initialize a plugin using these values
    Then the plugin is initialized correctly and all given values are set.
    """
    vcs = FakeVCS()
    plugin = FakePlugin(main_branch_name="sample_branch", vcs=vcs, check_result=True)

    assert plugin.main_branch_name == "sample_branch"
    assert plugin.vcs == vcs

    # Check if the dynamic property has been set
    assert plugin.check_result is True  # pylint: disable=no-member



from how_about_no.core import Config, Engine, Logger


def main() -> None:
    """Run the engine."""
    config = Config()
    Logger.initialize(config.logger_level)
    Engine(config).process()


if __name__ == "__main__":
    main()



import logging
from unittest.mock import Mock, PropertyMock, patch

import pytest

from how_about_no.tests.fakes import FakeEngine

MSG_PROCESSING_STARTED = "Processing started."
MSG_PROCESSING_FINISHED = "Processing finished."
MGS_CURRENT_BRANCH_IS_MAIN = "The current branch is the main repository branch."
MGS_BUILD_WILL_BE_SKIPPED = "The build will be skipped."
MGS_BUILD_WILL_NOT_BE_SKIPPED = "The build will not be skipped."
MSG_POPULATING_STARTED = "Populating the result using loaded plugins started."
MSG_POPULATING_FINISHED = "Populating the result using loaded plugins finished."
MGS_POPULATED_VALUE = "Populated value:"
MSG_CHECK_STARTED = "Checking all plugins for the output started."
MSG_CHECK_FINISHED = "Checking all plugins for the output finished."
MGS_CHECK_RESULT = "Check result:"
MGS_NO_PLUGINS_LOADED = "No plugins have been loaded, the build will not be skipped."


@patch("how_about_no.core.engine.load_vcs", Mock())
def test_current_branch_is_main_dont_skip_build(caplog, capsys, vcs, plugin, config):
    """
    Given the current branch of the repository is the main branch
    And we don't skip the builds on the main branch
    When we run `process` method
    Then the builds are not skipped
    And expected messages are printed.
    """
    caplog.clear()
    main_branch_name = "sample_branch"

    sample_config = config(main_branch=main_branch_name)
    sample_vcs = vcs()
    sample_plugin = plugin(main_branch_name=main_branch_name, vcs=sample_vcs)

    engine = FakeEngine(
        configuration=sample_config, vcs=sample_vcs, plugins=[sample_plugin]
    )

    with caplog.at_level(logging.DEBUG):
        engine.process()

    stdout, _ = capsys.readouterr()

    assert MSG_PROCESSING_STARTED in caplog.text
    assert MSG_PROCESSING_FINISHED in caplog.text
    assert MGS_CURRENT_BRANCH_IS_MAIN in caplog.text
    assert MGS_BUILD_WILL_NOT_BE_SKIPPED in caplog.text
    assert MGS_POPULATED_VALUE in stdout
    assert MGS_CHECK_RESULT not in stdout


@patch("how_about_no.core.engine.load_vcs", Mock())
def test_current_branch_is_main_skip_build(caplog, capsys, vcs, plugin, config):
    """
    Given the current branch of the repository is the main branch
    And we skip the builds on the main branch
    When we run `process` method
    Then the builds are skipped
    And expected messages are printed.
    """
    caplog.clear()
    main_branch_name = "sample_branch"

    sample_config = config(
        main_branch=main_branch_name, skip_builds_on_main_branch=True
    )
    sample_vcs = vcs()
    sample_plugin = plugin(main_branch_name=main_branch_name, vcs=sample_vcs)

    engine = FakeEngine(
        configuration=sample_config, vcs=sample_vcs, plugins=[sample_plugin]
    )

    with caplog.at_level(logging.DEBUG):
        engine.process()

    stdout, _ = capsys.readouterr()

    assert MSG_PROCESSING_STARTED in caplog.text
    assert MSG_PROCESSING_FINISHED in caplog.text
    assert MGS_CURRENT_BRANCH_IS_MAIN in caplog.text
    assert MGS_BUILD_WILL_BE_SKIPPED in caplog.text
    assert MGS_POPULATED_VALUE in stdout
    assert MGS_CHECK_RESULT not in stdout


@patch("how_about_no.core.engine.load_vcs", Mock())
def test_check_and_populate(caplog, capsys, vcs, plugin, config):
    """
    Given the current branch of the repository is not the main branch
    When we run `process` method
    Then the all plugins are used to check and populate the results
    And expected messages are printed.
    """
    caplog.clear()
    main_branch_name = "sample_branch"

    sample_config = config(main_branch=main_branch_name)
    sample_vcs = vcs(current_branch_name="another_branch")
    sample_plugin = plugin(main_branch_name=main_branch_name, vcs=sample_vcs)

    engine = FakeEngine(
        configuration=sample_config, vcs=sample_vcs, plugins=[sample_plugin]
    )

    with caplog.at_level(logging.DEBUG):
        engine.process()

    stdout, _ = capsys.readouterr()

    assert MSG_PROCESSING_STARTED in caplog.text
    assert MSG_PROCESSING_FINISHED in caplog.text
    assert MGS_CURRENT_BRANCH_IS_MAIN not in caplog.text
    assert MSG_POPULATING_STARTED in caplog.text
    assert MSG_POPULATING_FINISHED in caplog.text
    assert MGS_POPULATED_VALUE in stdout
    assert MSG_CHECK_STARTED in caplog.text
    assert MSG_CHECK_FINISHED in caplog.text
    assert MGS_CHECK_RESULT in stdout


@patch("how_about_no.core.engine.load_vcs", Mock())
def test_all_plugins_are_used(vcs, config, plugin_mock):
    """
    Given the current branch of the repository is not the main branch
    When we run `process` method
    Then the all plugins are used to check and populate the results.
    """
    main_branch_name = "sample_branch"

    sample_config = config(main_branch=main_branch_name)
    sample_vcs = vcs(current_branch_name="another_branch")
    plugins = [
        plugin_mock(main_branch_name=main_branch_name, vcs=vcs),
        plugin_mock(main_branch_name=main_branch_name, vcs=vcs),
        plugin_mock(main_branch_name=main_branch_name, vcs=vcs),
    ]

    engine = FakeEngine(configuration=sample_config, vcs=sample_vcs, plugins=plugins)

    engine.process()

    for plugin in plugins:
        plugin.check.assert_called_once()
        plugin.populate.assert_called_once()


@patch("how_about_no.core.engine.load_vcs", Mock())
def test_no_plugins_loaded(caplog, vcs, config):
    """
    Given no plugins have been loaded
    When we run `process` method
    Then the processing will not start
    And expected message is printed.
    """
    caplog.clear()

    engine = FakeEngine(configuration=config(), vcs=vcs(), plugins=[])

    with caplog.at_level(logging.DEBUG):
        engine.process()

    assert MGS_NO_PLUGINS_LOADED in caplog.text
    assert MSG_PROCESSING_STARTED not in caplog.text


@pytest.mark.parametrize(
    "result_policy, plugin_one_result, plugin_two_result, expected_result",
    (
        ("all", True, True, True),
        ("all", True, False, False),
        ("all", False, False, False),
        ("any", True, True, True),
        ("any", True, False, True),
        ("any", False, False, False),
        ("none", True, True, False),
        ("none", True, False, False),
        ("none", False, False, True),
    ),
)
@patch("how_about_no.core.engine.load_vcs", Mock())
def test_result_is_correct(  # pylint: disable=too-many-arguments
    vcs,
    config,
    plugin_mock,
    result_policy,
    plugin_one_result,
    plugin_two_result,
    expected_result,
):
    """
    Given the current branch of the repository is not the main branch
    And there are multiple plugins defined
    And the `check` method from each plugin returns defined value
    And the result policy is defined
    When we run `process` method
    Then all plugins populate the correct overall result.
    """
    main_branch_name = "sample_branch"

    sample_config = config(main_branch=main_branch_name, result_policy=result_policy)
    sample_vcs = vcs(current_branch_name="another_branch")
    plugins = [
        plugin_mock(
            main_branch_name=main_branch_name, vcs=vcs, result=plugin_one_result
        ),
        plugin_mock(
            main_branch_name=main_branch_name, vcs=vcs, result=plugin_two_result
        ),
    ]

    engine = FakeEngine(configuration=sample_config, vcs=sample_vcs, plugins=plugins)

    engine.process()

    for plugin in plugins:
        plugin.check.assert_called_once()
        plugin.populate.assert_called_once_with(expected_result)


@patch("how_about_no.core.engine.load_vcs", Mock())
def test_incorrect_result_policy(vcs, config, plugin_mock):
    """
    Given the current branch of the repository is not the main branch
    And there are multiple plugins defined
    And the value of `result_policy` is incorrect
    When we run `process` method
    Then a SystemExit error is raised.
    """
    main_branch_name = "sample_branch"

    sample_config = config(main_branch=main_branch_name)

    sample_vcs = vcs(current_branch_name="another_branch")
    plugins = [
        plugin_mock(main_branch_name=main_branch_name, vcs=sample_vcs),
        plugin_mock(main_branch_name=main_branch_name, vcs=sample_vcs),
    ]

    engine = FakeEngine(configuration=sample_config, vcs=sample_vcs, plugins=plugins)

    with patch(
        "how_about_no.core.config.Config.result_policy", new_callable=PropertyMock
    ) as mock_result_policy:
        mock_result_policy.return_value = "wrong_value"

        with pytest.raises(SystemExit):
            engine.process()



import pytest

from how_about_no.tests.base_tests import BasePluginTest


class TestTeamcityPlugin(BasePluginTest):
    """Test for the Teamcity plugin."""

    plugin_path = "how_about_no.plugins.ci.teamcity"
    plugin_class_name = "Teamcity"
    populate_implemented = True

    @pytest.mark.parametrize("value", (True, False))
    def test_populate_message_is_correct(self, vcs, capsys, value):
        """
        Given the value to populate with the plugin
        When we call `populate` method and pass that value as an argument
        Then a correct message will be printed.
        """
        plugin = self.module.PLUGIN("branch", vcs)
        plugin.populate(value)

        stdout, _ = capsys.readouterr()
        assert (
            f"##teamcity[setParameter name='env.SKIP_BUILDS' value='{value}']" in stdout
        )

    def test_custom_variable_name(self, vcs, capsys):
        """
        Given a custom variable name that should be used when populating the values
        When we call `populate` method
        Then a correct message with that variable name will be printed.
        """
        plugin = self.module.PLUGIN("branch", vcs, variable_name="example")
        plugin.populate(True)

        stdout, _ = capsys.readouterr()
        assert "##teamcity[setParameter name='example' value='True']" in stdout



from typing import List, Optional
from unittest.mock import Mock

import pytest

from how_about_no.core import Logger
from how_about_no.core.vcs import BaseVCS
from how_about_no.tests.data import SAMPLE_CONFIG_FILE_CONTENT
from how_about_no.tests.fakes import FakeConfig, FakePlugin, FakeVCS


@pytest.fixture(name="config_file_content")
def config_file_content_fixture():
    """Return sample content of a configuration file."""
    return SAMPLE_CONFIG_FILE_CONTENT


@pytest.fixture(name="vcs")
def vcs_fixture():
    """Sample VCS fixture."""

    def _inner(
        raise_validation_error: Optional[bool] = False,
        current_branch_name: Optional[str] = "sample_branch",
        changed_files: Optional[List[str]] = None,
    ):
        return FakeVCS(
            raise_validation_error=raise_validation_error,
            current_branch_name=current_branch_name,
            changed_files=changed_files,
        )

    return _inner


@pytest.fixture(name="plugin")
def plugin_fixture():
    """Sample plugin fixture."""

    def _inner(
        main_branch_name: str,
        vcs: BaseVCS,
        check_result: bool = True,
        is_check_implemented: bool = True,
        is_populate_implemented: bool = True,
        **kwargs,
    ):
        return FakePlugin(
            main_branch_name=main_branch_name,
            vcs=vcs,
            check_result=check_result,
            is_check_implemented=is_check_implemented,
            is_populate_implemented=is_populate_implemented,
            **kwargs,
        )

    return _inner


@pytest.fixture(name="plugin_mock")
def plugin_mock_fixture():
    """Sample plugin mock."""

    def _inner(main_branch_name: str, vcs: BaseVCS, result: Optional[bool] = True):
        mock = Mock(main_branch_name=main_branch_name, vcs=vcs)
        mock.check.return_value = result
        return mock

    return _inner


@pytest.fixture(name="config")
def config_fixture():
    """Sample config fixture."""

    def _inner(**kwargs):
        config = FakeConfig(**kwargs)
        Logger.initialize(config.logger_level)
        return config

    return _inner



import re
from typing import List

from how_about_no.core import BasePlugin


class ChangedFiles(BasePlugin):
    """Changed files plugin.

    If checks the list of all files changed compared to the main VCS branch against
    provided regex patterns. If all files match at least one of provided patterns,
    the result of the check is True, otherwise it's False.
    """

    def _get_regex_patterns(self):
        """
        Return the regex patterns that will be used to determine whether or not the
        build should be skipped.
        """
        return getattr(self, "patterns", [])

    def _does_file_path_match(  # pylint: disable=no-self-use
        self, file_path: str, patterns: List[str]
    ) -> bool:
        """Check if the name of the file match any of the provided regex patterns.

        Arguments:
            file_path (str): The path of the file to check.
            patterns (List[str]): A list of patterns to test.

        Returns:
            (bool): Flag indicating wether or not the name of the file match any of the
                provided regex patterns.
        """
        for pattern in patterns:
            if re.match(pattern, file_path):
                return True

        return False

    def check(self) -> bool:
        """Check if the files changed in the current branch can be ignored.

        Returns:
            (bool): Flag indicating whether or not the build should be skipped.
        """
        changed_files = self.vcs.get_list_of_changed_files(self.main_branch_name)
        if not changed_files:
            # If no files have been changed then we can skip the build.
            self._logger.debug("No files have been changed, the build can be skipped.")
            return True

        patterns = self._get_regex_patterns()
        if not patterns:
            # If there are no patterns provided then we cannot skip the build.
            self._logger.debug(
                "No regex patterns have been provided, the build cannot be skipped."
            )
            return False

        for file_path in changed_files:
            if not self._does_file_path_match(file_path, patterns):
                # If at least one files does not match any of the provided regex
                # patterns then we cannot skip the build.
                self._logger.debug(
                    f"File `{file_path}` does not match any of the provided regex"
                    " patterns."
                )
                return False

        # All files match at least one pattern, therefore we can skip the build.
        self._logger.debug(
            "All files match at least one of the provided regex patterns."
        )
        return True


PLUGIN = ChangedFiles



import logging

from how_about_no.tests.base_tests import BasePluginTest


class TestChangedFilesPlugin(BasePluginTest):
    """Test for the changed files plugin."""

    plugin_path = "how_about_no.plugins.vcs.changed_files"
    plugin_class_name = "ChangedFiles"
    check_implemented = True

    def test_check_no_changed_files(self, vcs, caplog):
        """
        Given no files have been changed comparing to the main branch
        When we call `check` method
        Then the result is True and a correct message is printed.
        """
        caplog.clear()

        plugin = self.module.PLUGIN("branch", vcs(changed_files=[]))

        with caplog.at_level(logging.DEBUG):
            result = plugin.check()

        assert result is True
        assert "No files have been changed, the build can be skipped." in caplog.text

    def test_check_no_patterns(self, vcs, caplog):
        """
        Given a plugin with no regex patterns configured
        When we call `check` method
        Then the result is False and a correct message is printed.
        """
        caplog.clear()

        plugin = self.module.PLUGIN(
            "branch", vcs(changed_files=["sample"]), patterns=[]
        )

        with caplog.at_level(logging.DEBUG):
            result = plugin.check()

        assert result is False
        assert (
            "No regex patterns have been provided, the build cannot be skipped."
            in caplog.text
        )

    def test_file_name_matches_patterns(self, vcs, caplog):
        """
        Given a file changed comparing to the main branch
        And a plugin regex pattern matching the name of the changed file
        When we call `check` method
        Then the result is True and a correct message is printed.
        """
        caplog.clear()

        plugin = self.module.PLUGIN(
            "branch", vcs(changed_files=["sample"]), patterns=["^sample$"]
        )

        with caplog.at_level(logging.DEBUG):
            result = plugin.check()

        assert result is True
        assert (
            "All files match at least one of the provided regex patterns."
            in caplog.text
        )

    def test_file_name_doesnt_match_patterns(self, vcs, caplog):
        """
        Given a file changed comparing to the main branch
        And a plugin regex pattern not matching the name of the changed file
        When we call `check` method
        Then the result is False and a correct message is printed.
        """
        caplog.clear()

        plugin = self.module.PLUGIN(
            "branch", vcs(changed_files=["sample"]), patterns=["^another$", "^pattern"]
        )

        with caplog.at_level(logging.DEBUG):
            result = plugin.check()

        assert result is False
        assert (
            "File `sample` does not match any of the provided regex patterns."
            in caplog.text
        )



import logging
from unittest.mock import patch

import pytest

from how_about_no.core.vcs import VCSException
from how_about_no.tests.base_tests import BaseVCSTest
from how_about_no.tests.fakes import get_fake_popen
from how_about_no.vcs.git import Git, get_git_executable_name


@pytest.mark.parametrize(
    "os_name, expected_result", (("posix", "git"), ("nt", "git.exe"))
)
def test_get_git_executable_name(os_name, expected_result):
    """
    Given the name of a operating system
    When we call get_git_executable_name
    Then we get a correct Git executable name.
    """
    with patch("how_about_no.vcs.git.os.name", os_name):
        assert get_git_executable_name() == expected_result


class TestGitVCSModule(BaseVCSTest):
    """Base tests for the Git VCS implementation."""

    vcs_path = "how_about_no.vcs.git"
    vcs_class_name = "Git"


class TestGetCurrentBranchName:
    """Tests for the Git `get_current_branch_name` method."""

    @pytest.fixture(autouse=True)
    def set_up(self):
        """Create a VCS object for each test."""
        self.vcs = Git()  # pylint: disable=attribute-defined-outside-init

    @patch("how_about_no.vcs.git.Popen", get_fake_popen(command_output="master\n"))
    def test_success(self):
        """
        Given git command returns a correct name of the branch
        When we call method `get_current_branch_name`
        Then we get the name of the branch as result.
        """
        assert self.vcs.get_current_branch_name() == "master"

    @patch("how_about_no.vcs.git.Popen", get_fake_popen(return_code=1))
    def test_incorrect_return_code(self, caplog):
        """
        Given git command returns a response with incorrect return code
        When we call method `get_current_branch_name`
        Then a VCSException error is raised and a correct error message is printed.
        """
        caplog.clear()

        with pytest.raises(VCSException):
            with caplog.at_level(logging.DEBUG):
                self.vcs.get_current_branch_name()

        assert (
            "It's not possible to fetch the name of the current branch." in caplog.text
        )

    @pytest.mark.parametrize("error", (IndexError, AttributeError))
    def test_raises_error(self, error, caplog):
        """
        Given executing the git command raises an error
        When we call method `get_current_branch_name`
        Then a VCSException error is raised and a correct error message is printed.
        """
        caplog.clear()

        with patch("how_about_no.vcs.git.Popen", get_fake_popen(side_effect=error)):
            with pytest.raises(VCSException):
                with caplog.at_level(logging.DEBUG):
                    self.vcs.get_current_branch_name()

        assert (
            "It's not possible to fetch the name of the current branch." in caplog.text
        )


class TestGetListOfChangedFiles:
    """Tests for the Git `get_list_of_changed_files` method."""

    @pytest.fixture(autouse=True)
    def set_up(self):
        """Create a VCS object for each test."""
        self.vcs = Git()  # pylint: disable=attribute-defined-outside-init

    @patch(
        "how_about_no.vcs.git.Popen", get_fake_popen(command_output="file1\nfile2\n")
    )
    def test_success(self):
        """
        Given git command returns a correct name of the branch
        When we call method `get_list_of_changed_files`
        Then we get the name of the branch as result.
        """
        assert self.vcs.get_list_of_changed_files("branch") == ["file1", "file2"]

    @patch("how_about_no.vcs.git.Popen", get_fake_popen(return_code=1))
    def test_incorrect_return_code(self, caplog):
        """
        Given git command returns a response with incorrect return code
        When we call method `get_list_of_changed_files`
        Then a VCSException error is raised and a correct error message is printed.
        """
        caplog.clear()

        with pytest.raises(VCSException):
            with caplog.at_level(logging.DEBUG):
                self.vcs.get_list_of_changed_files("branch")

        assert "It's not possible to fetch the list of changed files." in caplog.text

    @pytest.mark.parametrize("error", (IndexError, AttributeError))
    def test_raises_error(self, error, caplog):
        """
        Given executing the git command raises an error
        When we call method `get_list_of_changed_files`
        Then a VCSException error is raised and a correct error message is printed.
        """
        caplog.clear()

        with patch("how_about_no.vcs.git.Popen", get_fake_popen(side_effect=error)):
            with pytest.raises(VCSException):
                with caplog.at_level(logging.DEBUG):
                    self.vcs.get_list_of_changed_files("branch")

        assert "It's not possible to fetch the list of changed files." in caplog.text


class TestValidate:
    """Tests for the Git `validate` method."""

    @pytest.fixture(autouse=True)
    def set_up(self):
        """Create a VCS object for each test."""
        self.vcs = Git()  # pylint: disable=attribute-defined-outside-init

    @patch("how_about_no.vcs.git.Popen", get_fake_popen())
    @patch("how_about_no.vcs.git.which")
    def test_success(self, mock_which):
        """
        Given the git executable exists in the system PATH
        And the current working directory is a git repository
        When we call `validate` method
        Then it does not raise an error.
        """
        mock_which.return_value = "some_value"

        self.vcs.validate()

    @patch("how_about_no.vcs.git.Popen", get_fake_popen())
    @patch("how_about_no.vcs.git.which")
    def test_no_git_executable_found(self, mock_which):
        """
        Given the git executable does not exist in the system PATH
        When we call `validate` method
        Then it raises a VCSException error.
        """
        mock_which.return_value = None

        with pytest.raises(VCSException):
            self.vcs.validate()

    @patch("how_about_no.vcs.git.Popen", get_fake_popen(return_code=1))
    @patch("how_about_no.vcs.git.which")
    def test_no_git_repository_found(self, mock_which):
        """
        Given the git executable exists in the system PATH
        And the current working directory is not a git repository
        When we call `validate` method
        Then it raises a VCSException error.
        """
        mock_which.return_value = "some_value"

        with pytest.raises(VCSException):
            self.vcs.validate()



from unittest.mock import patch

from how_about_no.core import Logger


@patch("how_about_no.core.logger.logging")
def test_initialize(mock_logging):
    """
    When we call method `Logger.initialize`
    Then the logging `basicConfig` function is called with correct arguments.
    """
    Logger.initialize(level="sample")

    mock_logging.basicConfig.assert_called_once_with(
        level="sample", format="%(levelname)s - %(message)s"
    )


def test_get_existing_logger():
    """
    Given a logger with the given name already exists
    When we get a logger with the same name
    Then the same logger is returned.
    """
    logger = Logger.get_logger("sample")

    assert logger
    assert logger == Logger.get_logger("sample")


def test_get_new_logger():
    """
    Given a logger with the given name already exists
    When we get a logger with a different name
    Then a different logger is returned.
    """
    logger = Logger.get_logger("sample")

    assert logger
    assert logger != Logger.get_logger("another")



from how_about_no.core.config import Config
from how_about_no.core.engine import Engine
from how_about_no.core.logger import Logger
from how_about_no.core.plugin import BasePlugin
from how_about_no.core.vcs import BaseVCS, VCSException



import sys
from typing import List

from how_about_no.core.config import Config
from how_about_no.core.logger import Logger
from how_about_no.core.plugin import BasePlugin, load_plugin
from how_about_no.core.vcs import BaseVCS, load_vcs

RESULT_POLICY_FUNCTIONS = {
    "all": all,
    "any": any,
    "none": lambda values: all(not value for value in values),
}


class Engine:
    """
    Main class executing the checks and populating the value using handlers.
    """

    _plugins: List[BasePlugin] = None
    _vcs: BaseVCS = None

    def __init__(self, configuration: Config):
        """
        Validate whether or not the current working directory is a VCS repository and
        load all plugins defined in the configuration file.
        """
        self._logger = Logger.get_logger()
        self._configuration = configuration

        self._vcs = load_vcs(self._configuration.vcs)
        self._plugins = self._load_plugins(self._configuration.plugins)

    def _load_plugins(self, plugins: dict) -> List[BasePlugin]:
        """Load all plugins.

        Arguments:
            plugins (dict): A dictionary with all defined plugins. The key of each entry
                is the name of the plugin, the value is it's configuration.
        """
        self._logger.debug("Loading plugins started.")

        result = []

        for plugin_name, plugin_configuration in plugins.items():
            plugin = load_plugin(
                plugin_name,
                main_branch_name=self._configuration.main_branch,
                vcs=self._vcs,
                **plugin_configuration,
            )
            if not plugin:
                self._logger.warning("Plugin %s has not been loaded.", plugin_name)
                continue

            result.append(plugin)
            self._logger.debug("Plugin %s has been loaded.", plugin_name)

        self._logger.debug("Loading plugins finished.\n")

        return result

    def _is_current_main_branch(self) -> bool:
        """
        Checks whether or not the current branch in the repository is the main branch.
        """
        return self._vcs.get_current_branch_name() == self._configuration.main_branch

    def _check(self) -> bool:
        """Run the check using all loaded plugins.

        If all plugins decide that the build should be skipped then it will be.

        If the `check` method has not been implemented in a plugin then it will be
        skipped.

        Returns:
            (bool): The result of the check.
        """
        self._logger.debug("Checking all plugins for the output started.")
        result = []

        for plugin in self._plugins:
            try:
                result.append(plugin.check())
            except NotImplementedError:
                self._logger.debug(
                    "Method `check` has not been implemented in plugin %s. Skipping.",
                    plugin.__class__.__module__,
                )

        self._logger.debug("Checking all plugins for the output finished.\n")

        return result

    def _populate_result(self, result: bool) -> None:
        """Populate the result of the check using all loaded plugins.

        If the `populate` method has not been implemented in a plugin then it will be
        skipped.
        """
        self._logger.debug("Populating the result using loaded plugins started.")

        for plugin in self._plugins:
            try:
                plugin.populate(result)
            except NotImplementedError:
                self._logger.debug(
                    "Method `populate` has not been implemented in plugin %s."
                    " Skipping.",
                    plugin.__class__.__module__,
                )

        self._logger.debug("Populating the result using loaded plugins finished.\n")

    def _calculate_result(self, values: List[bool]) -> bool:
        """Calculate the final result.

        The result is calculated based on the result of method `check` from all plugins
        and seleted result policy.

        Arguments:
            values (List[bool]): The results from plugins.

        Returns:
            (bool)

        Raises:
            (SystemExit): When the value of the result policy is incorrect. This should
                not happen unless the Configuration class has been changed.
        """
        result_policy = self._configuration.result_policy
        result_policy_fun = RESULT_POLICY_FUNCTIONS.get(result_policy)
        if not result_policy_fun:
            self._logger.error("Incorrect result policy value: %s", result_policy)
            sys.exit(1)

        result = result_policy_fun(values)

        self._logger.info("The build will %sbe skipped.\n", "" if result else "not ")

        return result

    def process(self) -> None:
        """Run the check using all loaded plugins and optional configuration file."""
        if not self._plugins:
            self._logger.warning(
                "No plugins have been loaded, the build will not be skipped."
            )
            return

        self._logger.debug("Processing started.\n")

        if self._is_current_main_branch():
            # Skip the builds on the main branch if the configuration says so
            result = self._configuration.skip_builds_on_main_branch
            self._logger.debug(
                "The current branch is the main repository branch. The build will %s"
                "be skipped.",
                "" if result else "not ",
            )
        else:
            # Fetch the result based on the output of all plugins
            result = self._calculate_result(self._check())

        # Populate the output using all plugins.
        self._populate_result(result)

        self._logger.debug("Processing finished.")



import logging
import os
from unittest.mock import mock_open, patch

import pytest

from how_about_no.core import Config
from how_about_no.tests.data import SAMPLE_CONFIG_DICT


class TestLoadingConfiguration:
    """Tests for the core configuration module."""

    @patch.dict(os.environ, {"HOW_ABOUT_NO_CONFIG_FILE": ""}, clear=True)
    @patch("how_about_no.core.config.Config._load")
    def test_default_config_file_used(self, mock_load):  # pylint: disable=no-self-use
        """
        Given no custom path to the configuration file is set
        When we load the configuration
        Then the default path to the config file is used.
        """
        default_path = f"{os.getcwd()}/how_about_no.yaml"

        Config()

        mock_load.assert_called_once_with(default_path)

    @patch.dict(os.environ, {"HOW_ABOUT_NO_CONFIG_FILE": "SAMPLE_PATH"}, clear=True)
    @patch("how_about_no.core.config.os.path.exists")
    @patch("how_about_no.core.config.Config._load")
    def test_custom_non_existing_config_file_used(  # pylint: disable=no-self-use
        self, mock_load, mock_exists, caplog
    ):
        """
        Given a custom path to an existing configuration file is set
        When we load the configuration
        Then the custom path to the config file is used
        And a ValueError is raised.
        """
        mock_exists.return_value = False

        with caplog.at_level(logging.WARNING):
            with pytest.raises(ValueError):
                Config()

        mock_exists.assert_called_once_with("SAMPLE_PATH")
        mock_load.assert_not_called()

    @patch.dict(os.environ, {"HOW_ABOUT_NO_CONFIG_FILE": "SAMPLE_PATH"}, clear=True)
    @patch("how_about_no.core.config.os.path.exists")
    @patch("how_about_no.core.config.Config._load")
    def test_custom_existing_config_file_used(  # pylint: disable=no-self-use
        self, mock_load, mock_exists
    ):
        """
        Given a custom path to an existing configuration file is set
        When we load the configuration
        Then the custom path to the config file is used.
        """
        mock_exists.return_value = True

        Config()

        mock_exists.assert_called_once_with("SAMPLE_PATH")
        mock_load.assert_called_once_with("SAMPLE_PATH")

    @patch.dict(os.environ, {"HOW_ABOUT_NO_CONFIG_FILE": ""}, clear=True)
    @patch("how_about_no.core.config.os.path.exists")
    def test_config_loaded_correctly(  # pylint: disable=no-self-use
        self, mock_exists, config_file_content
    ):
        """
        Given the `open` function returns a file with some content
        When we load the configuration using that file
        Then the configuration is loaded and contains the data from that file.
        """
        mock_exists.return_value = True

        with patch("builtins.open", mock_open(read_data=config_file_content)):
            config = Config()

        assert config.configuration == SAMPLE_CONFIG_DICT


class TestMainBranchProperty:
    """Tests for property `main_branch`."""

    @patch("how_about_no.core.config.os.path.exists")
    @patch("how_about_no.core.config.Config._load")
    @pytest.mark.parametrize("configuration_value", (1, "abc", tuple(), set()))
    def test_with_configuration_not_set_to_dict(  # pylint: disable=no-self-use
        self, mock_load, mock_exists, configuration_value
    ):
        """
        Given the configuration object set to a certain value
        When we fetch the main branch using `main_branch` property
        Then we get a correct value as the result.
        """
        mock_exists.return_value = True
        mock_load.return_value = configuration_value

        config = Config()

        assert config.main_branch == "master"

    @patch("how_about_no.core.config.os.path.exists")
    @patch("how_about_no.core.config.Config._load")
    @pytest.mark.parametrize(
        "configuration_value, expected_result",
        (
            (1, "master"),
            ("abc", "abc"),
            (tuple(), "master"),
            ({}, "master"),
            (set(), "master"),
            (True, "master"),
            (False, "master"),
        ),
    )
    def test_with_configuration_set_to_dict(  # pylint: disable=no-self-use
        self, mock_load, mock_exists, configuration_value, expected_result
    ):
        """
        Given the `main_branch` setting set to a certain value
        When we fetch the main branch using `main_branch` property
        Then we get a correct value as the result.
        """
        mock_exists.return_value = True
        mock_load.return_value = {"main_branch": configuration_value}

        config = Config()

        assert config.main_branch == expected_result


class TestSkipBuildsOnMainBranchProperty:
    """Tests for property `skip_builds_on_main_branch`."""

    @patch("how_about_no.core.config.os.path.exists")
    @patch("how_about_no.core.config.Config._load")
    @pytest.mark.parametrize("configuration_value", (1, "abc", tuple(), set()))
    def test_with_configuration_not_set_to_dict(  # pylint: disable=no-self-use
        self, mock_load, mock_exists, configuration_value
    ):
        """
        Given the configuration object set to a certain value
        When we fetch the main branch using `skip_builds_on_main_branch` property
        Then we get a correct value as the result.
        """
        mock_exists.return_value = True
        mock_load.return_value = configuration_value

        config = Config()

        assert config.skip_builds_on_main_branch is False

    @patch("how_about_no.core.config.os.path.exists")
    @patch("how_about_no.core.config.Config._load")
    @pytest.mark.parametrize(
        "configuration_value, expected_result",
        (
            (1, False),
            ("abc", False),
            (tuple(), False),
            ({}, False),
            (set(), False),
            (True, True),
            (False, False),
        ),
    )
    def test_with_configuration_set_to_dict(  # pylint: disable=no-self-use
        self, mock_load, mock_exists, configuration_value, expected_result
    ):
        """
        Given the `skip_builds_on_main_branch` setting set to a certain value
        When we fetch the main branch using `skip_builds_on_main_branch` property
        Then we get a correct value as the result.
        """
        mock_exists.return_value = True
        mock_load.return_value = {"skip_builds_on_main_branch": configuration_value}

        config = Config()

        assert config.skip_builds_on_main_branch == expected_result


class TestLoggerLevelProperty:
    """Tests for property `logger_level`."""

    @patch("how_about_no.core.config.os.path.exists")
    @patch("how_about_no.core.config.Config._load")
    @pytest.mark.parametrize("configuration_value", (1, "abc", tuple(), set()))
    def test_with_configuration_not_set_to_dict(  # pylint: disable=no-self-use
        self, mock_load, mock_exists, configuration_value
    ):
        """
        Given the configuration object set to a certain value
        When we fetch the main branch using `logger_level` property
        Then we get a correct value as the result.
        """
        mock_exists.return_value = True
        mock_load.return_value = configuration_value

        config = Config()

        assert config.logger_level == "DEBUG"

    @patch("how_about_no.core.config.os.path.exists")
    @patch("how_about_no.core.config.Config._load")
    @pytest.mark.parametrize(
        "configuration_value, expected_result",
        (
            (1, "DEBUG"),
            ("abc", "abc"),
            (tuple(), "DEBUG"),
            ({}, "DEBUG"),
            (set(), "DEBUG"),
            (True, "DEBUG"),
            (False, "DEBUG"),
        ),
    )
    def test_with_configuration_set_to_dict(  # pylint: disable=no-self-use
        self, mock_load, mock_exists, configuration_value, expected_result
    ):
        """
        Given the `logger_level` setting set to a certain value
        When we fetch the main branch using `logger_level` property
        Then we get a correct value as the result.
        """
        mock_exists.return_value = True
        mock_load.return_value = {"logger_level": configuration_value}

        config = Config()

        assert config.logger_level == expected_result


class TestPluginsProperty:
    """Tests for property `plugins`."""

    @patch("how_about_no.core.config.os.path.exists")
    @patch("how_about_no.core.config.Config._load")
    @pytest.mark.parametrize("configuration_value", (1, "abc", tuple(), set()))
    def test_with_configuration_not_set_to_dict(  # pylint: disable=no-self-use
        self, mock_load, mock_exists, configuration_value
    ):
        """
        Given the configuration object set to a certain value
        When we fetch the main branch using `plugins` property
        Then we get a correct value as the result.
        """
        mock_exists.return_value = True
        mock_load.return_value = configuration_value

        config = Config()

        assert config.plugins == {}

    @patch("how_about_no.core.config.os.path.exists")
    @patch("how_about_no.core.config.Config._load")
    @pytest.mark.parametrize(
        "configuration_value, expected_result",
        (
            (1, {}),
            ("abc", {}),
            (tuple(), {}),
            ({"sample": "value"}, {"sample": "value"}),
            (set(), {}),
            (True, {}),
            (False, {}),
        ),
    )
    def test_with_configuration_set_to_dict(  # pylint: disable=no-self-use
        self, mock_load, mock_exists, configuration_value, expected_result
    ):
        """
        Given the `plugins` setting set to a certain value
        When we fetch the main branch using `plugins` property
        Then we get a correct value as the result.
        """
        mock_exists.return_value = True
        mock_load.return_value = {"plugins": configuration_value}

        config = Config()

        assert config.plugins == expected_result


class TestResultPolicyProperty:
    """Tests for property `result_policy`."""

    @patch("how_about_no.core.config.os.path.exists")
    @patch("how_about_no.core.config.Config._load")
    @pytest.mark.parametrize("configuration_value", (1, "abc", tuple(), set()))
    def test_with_configuration_not_set_to_dict(  # pylint: disable=no-self-use
        self, mock_load, mock_exists, configuration_value
    ):
        """
        Given the configuration object set to a certain value
        When we fetch the main branch using `result_policy` property
        Then we get a correct value as the result.
        """
        mock_exists.return_value = True
        mock_load.return_value = configuration_value

        config = Config()

        assert config.result_policy == "all"

    @patch("how_about_no.core.config.os.path.exists")
    @patch("how_about_no.core.config.Config._load")
    @pytest.mark.parametrize(
        "configuration_value, expected_result",
        (
            (1, "all"),
            ("abc", "all"),
            ("all", "all"),
            ("any", "any"),
            ("none", "none"),
            (tuple(), "all"),
            ({}, "all"),
            (set(), "all"),
            (True, "all"),
            (False, "all"),
        ),
    )
    def test_with_configuration_set_to_dict(  # pylint: disable=no-self-use
        self, mock_load, mock_exists, configuration_value, expected_result
    ):
        """
        Given the `result_policy` setting set to a certain value
        When we fetch the main branch using `result_policy` property
        Then we get a correct value as the result.
        """
        mock_exists.return_value = True
        mock_load.return_value = {"result_policy": configuration_value}

        config = Config()

        assert config.result_policy == expected_result



import importlib
from abc import ABC, abstractmethod
from typing import List

from how_about_no.core.logger import Logger


class VCSException(Exception):
    """
    This is an exception that should be raised in case the VCS validation fails.
    """


class BaseVCS(ABC):
    """This is a base class for all Version Control System implementations.

    In order to keep this library generic, it should be possible to use it
    against different Version Control Systems (VCSs). This class provides an
    interface that should be implemented by all child classes and can be used
    in the plugins.
    """

    def __init__(self):
        self._logger = Logger.get_logger()

    @abstractmethod
    def validate(self) -> None:
        """
        Validate whether or not the current repository can be managed by given
        VCS implementation.

        If the validation fails, VCSException should be raised.
        """

    @abstractmethod
    def get_current_branch_name(self) -> str:
        """Return the name of the branch that is currently checked out."""

    @abstractmethod
    def get_list_of_changed_files(self, branch_name: str) -> List[str]:
        """Return a list of all files changed comparing to given branch."""


def load_vcs(vcs_module_path: str) -> BaseVCS:
    """Load and return the version control system from given module.

    Arguments:
        vcs_module_path (str): The path of the VCS module.

    Returns:
        (BaseVCS)

    Raises:
        (VCSException): When it's not possible to load the VCS.
    """
    try:
        return importlib.import_module(vcs_module_path).VCS()
    except (AttributeError, ImportError, TypeError) as ex:
        raise VCSException("Given version control system could not be loaded.") from ex



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['how_about_no',
 'how_about_no.core',
 'how_about_no.plugins',
 'how_about_no.plugins.ci',
 'how_about_no.plugins.vcs',
 'how_about_no.tests',
 'how_about_no.tests.test_core',
 'how_about_no.tests.test_plugins',
 'how_about_no.tests.test_plugins.test_ci',
 'how_about_no.tests.test_plugins.test_vcs',
 'how_about_no.tests.test_vcs',
 'how_about_no.vcs']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.4.1,<6.0.0']

entry_points = \
{'console_scripts': ['how-about-no = how_about_no:main']}

setup_kwargs = {
    'name': 'how-about-no',
    'version': '0.1.0b1',
    'description': "Let's your CI/CD tool know when the build can be skipped.",
    'long_description': '# How about no\n\n![tests][tests]\n![black][black]\n![pylint][pylint]\n![isort][isort]\n\n![how_about_no][how_about_no]\n\n*Image source: https://knowyourmeme.com/photos/129577-how-about-no*\n\nThe sole purpose of this library is to let your CI/CD tool know whether or not the\nbuild it\'s about to run is necessary.\n\nPlease note that there is no stable release yet, and the documentation will be added\nshortly.\n\n\n\n[tests]: https://github.com/mateuszcisek/how_about_no/actions/workflows/tests.yaml/badge.svg "tests"\n[black]: https://github.com/mateuszcisek/how_about_no/actions/workflows/linting_black.yaml/badge.svg "black"\n[pylint]: https://github.com/mateuszcisek/how_about_no/actions/workflows/linting_pylint.yaml/badge.svg "pylint"\n[isort]: https://github.com/mateuszcisek/how_about_no/actions/workflows/linting_isort.yaml/badge.svg "isort"\n[how_about_no]: how_about_no.jpg "how about no"\n',
    'author': 'Mateusz Cisek',
    'author_email': 'cismat@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mateuszcisek/how_about_no',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import logging
from typing import Dict, Optional


class Logger:
    """Wrapper for the logger."""

    _instances: Dict[str, logging.Logger] = {}

    @staticmethod
    def initialize(level: str) -> None:
        """Initialize the logger with given log level.

        Once this is done, the log can be fetched using `get_logger` method.

        Arguments:
            level (str): The level to set in the logger.
        """
        logging.basicConfig(level=level, format="%(levelname)s - %(message)s")

    @staticmethod
    def get_logger(name: Optional[str] = "app") -> logging.Logger:
        """
        Get the instance of the logger with given name.

        Arguments:
            name (str): The name of the logger to fetch.

        Returns:
            (logging.Logger)
        """
        if name not in Logger._instances:
            Logger._instances[name] = logging.getLogger(name)

        return Logger._instances[name]



import sys

from how_about_no.core import BasePlugin


class Teamcity(BasePlugin):
    """Teamcity plugin.

    It is used to populate the result of the check to a Teamcity variable.
    """

    def _get_variable_name(self) -> str:
        """Return the name of the variable to set on Teamcity."""
        return str(getattr(self, "variable_name", "env.SKIP_BUILDS"))

    def populate(self, value: bool) -> None:
        """Set the value of a Teamcity variable to the given value.

        Arguments:
            value (bool): The value to set.
        """
        variable = self._get_variable_name()
        sys.stdout.write(
            f"##teamcity[setParameter name='{variable}' value='{value}']\n"
        )


PLUGIN = Teamcity



import importlib
from abc import ABC

import pytest


class BasePluginTest(ABC):
    """Base class for testing the plugins."""

    plugin_path: str = None
    plugin_class_name: str = None
    check_implemented: bool = False
    populate_implemented: bool = False

    @pytest.fixture(autouse=True)
    def set_up(self):
        """Import the module before each test."""
        # pylint: disable=attribute-defined-outside-init
        self.module = importlib.import_module(self.plugin_path)

    def test_plugin_exported_correctly(self):
        """
        Given the path to the plugin module and the name of the plugin class
        When we import both plugin class and `PLUGIN` constant from the module
        Then both plugin class and `PLUGIN` object are the same.
        """
        assert getattr(self.module, self.plugin_class_name) == getattr(
            self.module, "PLUGIN"
        )

    def test_check_is_implemented(self, vcs):
        """
        Given the fact that the `check` method is or is not implemented in the plugin
        When we run `check` method from the plugin
        Then a NotImplementedError is or is not raised.
        """
        plugin = self.module.PLUGIN("branch", vcs())

        if self.check_implemented:
            plugin.check()
        else:
            with pytest.raises(NotImplementedError):
                plugin.check()

    def test_populate_is_implemented(self, vcs):
        """
        Given the fact that the `populate` method is or is not implemented in the plugin
        When we run `populate` method from the plugin
        Then a NotImplementedError is or is not raised.
        """
        plugin = self.module.PLUGIN("branch", vcs())

        if self.populate_implemented:
            plugin.populate(True)
        else:
            with pytest.raises(NotImplementedError):
                plugin.populate(True)


class BaseVCSTest(ABC):
    """Base class for testing the VCS implementations."""

    vcs_path: str = None
    vcs_class_name: str = None

    @pytest.fixture(autouse=True)
    def set_up(self):
        """Import the module before each test."""
        # pylint: disable=attribute-defined-outside-init
        self.module = importlib.import_module(self.vcs_path)

    def test_vcs_exported_correctly(self):
        """
        Given the path to the VCS module and the name of the VCS class
        When we import both VCS class and `VCS` constant from the module
        Then both VCS class and `VCS` object are the same.
        """
        assert getattr(self.module, self.vcs_class_name) == getattr(self.module, "VCS")



import os
from typing import Any, Dict, Union

import yaml


class Config:
    """Class that holds the configuration from the configuration file."""

    _default_path: str = os.path.join(os.getcwd(), "how_about_no.yaml")
    _env_variable_name: str = "HOW_ABOUT_NO_CONFIG_FILE"
    _configuration: dict = None

    def __init__(self) -> None:
        """Load the configuration from the file.

        The default path of the configuration file is `./how_about_no.yaml`. If the
        default configuration file does not exist then the default settings will be
        used.

        If the environment variable `HOW_ABOUT_NO_CONFIG_FILE` is set, that path will be
        used instead. In that case, a ValueError error will be raised if the
        configuration file does not exist.

        Why it behaves like this? Well, it should be possible to use the library without
        defining the configuration file, so if it's not there then it's fine and we can
        expect to get the default values in that case. On the other hand, if there is a
        custom config file path defined then we assume the user wants to use his custom
        file and we should let him know that the provided path is incorrect if the
        specified file does not exist.
        """
        custom_path = os.getenv(self._env_variable_name)
        if custom_path:
            if not os.path.exists(custom_path):
                raise ValueError(
                    "Defined path to the configuration file is incorrect "
                    f"({custom_path}). Please make sure the file exists and "
                    "try again."
                )

            self._configuration = self._load(custom_path)
        else:
            if os.path.exists(self._default_path):
                self._configuration = self._load(self._default_path)

    def _load(self, file_path: str) -> Any:  # pylint: disable=no-self-use
        """Load and return the configuration from given file.

        Arguments:
            file_path (str): The path of the configuration file.
        """
        try:
            with open(file_path, encoding="utf-8") as config_file:
                return yaml.load(config_file, yaml.Loader)
        except Exception:  # pylint: disable=broad-except
            # If loading the configuration fails for any reason, return None.
            return None

    @property
    def configuration(self) -> Union[Dict, None]:
        """Return the configuration object."""
        return self._configuration

    @property
    def main_branch(self) -> str:
        """Return the value of the `main_branch` setting.

        If it's not defined or it's not a string then the default value will be
        returned.

        Default value: master.
        """
        default_value = "master"

        if not isinstance(self._configuration, dict):
            return default_value

        value = self._configuration.get("main_branch")
        return value if isinstance(value, str) else default_value

    @property
    def skip_builds_on_main_branch(self) -> bool:
        """Return the value of the `skip_builds_on_main_branch` setting.

        If it's not defined or it's not a boolean then the default value will be
        returned.

        Default value: False.
        """
        default_value = False

        if not isinstance(self._configuration, dict):
            return default_value

        value = self._configuration.get("skip_builds_on_main_branch")
        return value if isinstance(value, bool) else default_value

    @property
    def logger_level(self) -> str:
        """Return the value of the `logger_level` setting.

        If it's not defined or it's not a string then the default value will be
        returned.

        Default value: DEBUG.
        """
        default_value = "DEBUG"

        if not isinstance(self._configuration, dict):
            return default_value

        value = self._configuration.get("logger_level")
        return value if isinstance(value, str) else default_value

    @property
    def plugins(self) -> Dict:
        """Return the value of the `plugins` setting.

        If it's not defined or it's not a dictionary then the default value will be
        returned.

        Default value: {} (empty dictionary).
        """
        default_value = {}

        if not isinstance(self._configuration, dict):
            return default_value

        value = self._configuration.get("plugins")
        return value if isinstance(value, dict) else default_value

    @property
    def vcs(self) -> str:
        """Return the value of the `vcs` setting.

        If it's not defined or it's not a string then the default value will be
        returned.

        Default value: master.
        """
        default_value = "how_about_no.vcs.git"

        if not isinstance(self._configuration, dict):
            return default_value

        value = self._configuration.get("vcs")
        return value if isinstance(value, str) else default_value

    @property
    def result_policy(self) -> str:
        """Return the value of the `result_policy` setting.

        If it's not defined or it's not one of available values then the default value
        will be returned.

        Default value: all.
        """
        default_value = "all"

        if not isinstance(self._configuration, dict):
            return default_value

        value = self._configuration.get("result_policy")
        return value if value in ["all", "any", "none"] else default_value



from typing import List, Optional, Type
from unittest.mock import Mock

from how_about_no.core import BasePlugin, BaseVCS, Config, Engine, VCSException
from how_about_no.tests.data import SAMPLE_CONFIG_DICT
from how_about_no.vcs.git import Git


class FakeVCS(BaseVCS):
    """A fake VCS implementation for testing purposes."""

    def __init__(
        self,
        raise_validation_error: Optional[bool] = False,
        current_branch_name: Optional[str] = "sample_branch",
        changed_files: Optional[List[str]] = None,
    ):
        super().__init__()
        self.raise_validation_error = raise_validation_error
        self.current_branch_name = current_branch_name
        self.changed_files = changed_files

    def validate(self) -> None:
        if self.raise_validation_error:
            raise VCSException()

    def get_current_branch_name(self) -> str:
        return self.current_branch_name

    def get_list_of_changed_files(self, branch_name: str) -> List[str]:
        return self.changed_files if self.changed_files else []


class FakePlugin(BasePlugin):
    """A fake plugin implementation for testing purposes."""

    def __init__(
        self,
        main_branch_name: str,
        vcs: BaseVCS,
        is_check_implemented: bool = True,
        is_populate_implemented: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(main_branch_name, vcs, **kwargs)
        self._is_check_implemented = is_check_implemented
        self._is_populate_implemented = is_populate_implemented

    def check(self):
        if not self._is_check_implemented:
            raise NotImplementedError()

        result = getattr(self, "check_result", False)
        print(f"Check result: {result}")
        return result

    def populate(self, value: bool) -> None:
        if not self._is_populate_implemented:
            raise NotImplementedError()

        print(f"Populated value: {value}")


class FakeConfig(Config):
    """A fake configuration implementation for testing purposes."""

    def __init__(self, **kwargs) -> None:  # pylint: disable=super-init-not-called
        self._configuration = SAMPLE_CONFIG_DICT
        self._configuration.update(kwargs)


class FakeEngine(Engine):
    """A fake engine implementation for testing purposes."""

    def __init__(
        self, configuration: Config, vcs: BaseVCS, plugins: List[BasePlugin] = None
    ):
        self._defined_plugins = plugins or []
        super().__init__(configuration)

        self._vcs = vcs or Git()

    def _load_plugins(self, plugins: dict) -> List[BasePlugin]:
        return self._defined_plugins


def get_fake_popen(
    command_output: Optional[str] = None,
    return_code: Optional[int] = 0,
    side_effect: Optional[Type[Exception]] = None,
):
    """Return a fake class imitating the behavior of Popen class."""

    class _FakePopen:
        """A fake Popen implementation for testing purposes."""

        def __init__(self, *args, **kwargs):
            pass

        def __enter__(self):
            if side_effect:
                raise side_effect()

            mock_popen_result = Mock(returncode=return_code)
            mock_popen_result.communicate.return_value = (
                command_output.encode("utf-8") if command_output else "",
                None,
            )
            return mock_popen_result

        def __exit__(self, *args, **kwargs):
            pass

    return _FakePopen



import os
from shutil import which
from subprocess import PIPE, Popen
from typing import List

from how_about_no.core import BaseVCS, VCSException


def get_git_executable_name():
    """Return the name of the Git executable file."""
    return "git" if os.name == "posix" else "git.exe"


class Git(BaseVCS):
    """Git VCS implementation."""

    def __init__(self):
        super().__init__()

        self._exec_name = get_git_executable_name()

    def get_current_branch_name(self) -> str:
        """Return the name of the current Git branch.

        Returns:
            (str): The name of the current branch.

        Raises:
            (VCSException): When it's not possible to fetch the name
                of the current branch.
        """
        error_msg = "It's not possible to fetch the name of the current branch."

        try:
            with Popen(
                [self._exec_name, "branch", "--show-current"], stdout=PIPE
            ) as child:
                output = child.communicate()
                if child.returncode == 0:
                    return output[0].decode("utf-8").strip()

            self._logger.error(error_msg)
            raise VCSException(error_msg)
        except (IndexError, AttributeError) as error:
            self._logger.error(error_msg)
            raise VCSException(error_msg) from error

    def get_list_of_changed_files(self, branch_name: str) -> List[str]:
        """
        Return the list of all files that changed compared to a branch with the given
        name.

        Arguments:
            branch_name (str): The name of the branch to compare.

        Returns:
            (str): The name of the current branch.

        Raises:
            (VCSException): When it's not possible to fetch the list of changed
                files.
        """
        error_msg = "It's not possible to fetch the list of changed files."

        try:
            with Popen(
                [self._exec_name, "diff", branch_name, "--name-only"], stdout=PIPE
            ) as child:
                output = child.communicate()
                if child.returncode == 0:
                    return output[0].decode("utf-8").strip().split("\n")

            self._logger.error(error_msg)
            raise VCSException(error_msg)
        except (IndexError, AttributeError) as error:
            self._logger.error(error_msg)
            raise VCSException(error_msg) from error

    def _is_git_executable_available(self) -> bool:
        """Check whether or not the Git executable exists in the system."""
        return which(self._exec_name) is not None

    def _is_current_directory_a_git_repository(self) -> bool:
        """Check if the current directory is a Git repository."""
        with Popen(
            [self._exec_name, "rev-parse", "--is-inside-work-tree"], stdout=PIPE
        ) as child:
            child.communicate()
            return child.returncode == 0

    def validate(self) -> None:
        """
        Check if the `git` executable is reachable and that the working directory holds
        a Git repository.
        """
        if not self._is_git_executable_available():
            raise VCSException("Git executable is not available in the system PATH.")

        if not self._is_current_directory_a_git_repository():
            raise VCSException(
                "Current directory does not seem to be a Git repository."
            )


VCS = Git



SAMPLE_CONFIG_FILE_CONTENT = "\n".join(
    [
        "main_branch: sample_branch",
        "skip_builds_on_main_branch: false",
        "logger_level: ERROR",
        "vcs: vcs.git",
        "result_policy: all",
        "plugins:",
        "  sample_plugin:",
        "    option_1: text",
        "    option_2: 123",
    ]
)

SAMPLE_CONFIG_DICT = {
    "main_branch": "sample_branch",
    "skip_builds_on_main_branch": False,
    "logger_level": "ERROR",
    "vcs": "vcs.git",
    "result_policy": "all",
    "plugins": {"sample_plugin": {"option_1": "text", "option_2": 123}},
}
